import binascii

import nacl.utils

from semux.key import (
    Key,
    verify_message,
    PUBLIC_KEY_LEN,
    PRIVATE_KEY_LEN,
    ENCODED_PUBLIC_KEY_LEN,
    ENCODED_PRIVATE_KEY_LEN,
    ADDRESS_LEN
)
from semux.hash import hash256
from semux.signature import SIG_LEN, Signature

TEST_ENCODED_PRIVATEKEY = b'302e020100300506032b657004220420bd2f24b259aac4bfce3792c31d0f62a7f28b439c3e4feb97050efe5fe254f2af'
TEST_ENCODED_PUBLICKEY = b'302a300506032b6570032100b72dc8ebc9f53d21837dc96483da08765ea11f25c1bd4c3cb49318c944d67b9b'
TEST_ADDRESS = b'0680a919c78faa59b127014b6181979ae0a62dbd'
TEST_SIG = b'e345ba4cbdc7dcca559d7b60a4740a750d565409481ffb6ee18e4d0c104bdd8d37457e383b1cd515e2f8e1a4a3ecbd771a29ef0520e4b017ca6e005881f07606b72dc8ebc9f53d21837dc96483da08765ea11f25c1bd4c3cb49318c944d67b9b'


def test_key_constants():
    key = Key.from_random()
    assert len(key.public.encode()) == PUBLIC_KEY_LEN
    assert len(key.private.encode()) == PRIVATE_KEY_LEN
    assert len(key.encoded_public) == ENCODED_PUBLIC_KEY_LEN
    assert len(key.encoded_private) == ENCODED_PRIVATE_KEY_LEN
    assert len(key.to_address(True)) == ADDRESS_LEN


def test_from_encoded():
    key = Key.from_encoded_private(binascii.unhexlify(TEST_ENCODED_PRIVATEKEY))
    assert binascii.hexlify(key.encoded_private) == TEST_ENCODED_PRIVATEKEY
    assert binascii.hexlify(key.encoded_public) == TEST_ENCODED_PUBLICKEY
    assert key.to_address() == TEST_ADDRESS

    key2 = Key.from_encoded_private(key.encoded_private)
    assert key2.encoded_private == key.encoded_private
    assert key2.private == key.private
    assert key2.public == key.public
    assert key2.seed == key.seed
    assert key2.to_address(False) == key.to_address(False)
    assert key2.to_address() == key.to_address()


def test_sign_verify():
    msg = hash256(b'test')
    key = Key.from_random()
    sig = key.sign(msg)

    assert len(sig.to_bytes()) == SIG_LEN
    assert verify_message(msg, sig) is True
    assert sig.public_key == key.public.encode()
    assert sig.get_address() == key.to_address(True)


def test_sign_verify_large():
    msg = nacl.utils.random(1024 * 1024)
    key = Key.from_random()
    sig = key.sign(msg)

    assert verify_message(msg, sig) is True
    assert sig.public_key == key.public.encode()


def test_bad_signature():
    msg = hash256(b'test')
    random_sig = Signature.from_bytes(nacl.utils.random(SIG_LEN))

    assert verify_message(msg, random_sig) is False


def test_expected_sign():
    key = Key.from_encoded_private(binascii.unhexlify(TEST_ENCODED_PRIVATEKEY))
    sig = Signature.from_bytes(binascii.unhexlify(TEST_SIG))

    assert sig.public_key == key.public.encode()
    assert sig.signed_message == key.sign(b'test').signed_message
    assert sig.get_address() == key.to_address(True)
    assert sig.to_bytes() == binascii.unhexlify(TEST_SIG)



import binascii

import nacl.utils
import nacl.signing
import nacl.encoding
from nacl.exceptions import BadSignatureError

from .hash import hash160
from .encoding import to_pkcs8, from_pkcs8, to_x509
from .signature import Signature

__all__ = [
    "PUBLIC_KEY_LEN",
    "ENCODED_PUBLIC_KEY_LEN",
    "PRIVATE_KEY_LEN",
    "ENCODED_PRIVATE_KEY_LEN",
    "ADDRESS_LEN",
    "SEED_LEN",
    "verify_message",
    "Key",
]

PUBLIC_KEY_LEN = 32
ENCODED_PUBLIC_KEY_LEN = 44

PRIVATE_KEY_LEN = 32
ENCODED_PRIVATE_KEY_LEN = 48

SEED_LEN = 32
ADDRESS_LEN = 20


def verify_message(message, signature):
    """
    Verify a signed message.

    Args:
        message (bytes): plain text message
        signature (bytes): generated signature

    Returns:
        boolean. True if signature is valid, False otherwise.
    """
    public = signature.get_verifying_key()
    try:
        public.verify(message, signature.signed_message)
        return True
    except BadSignatureError:
        return False


def generate_keypair(seed=None):
    """
    Generate a new key pair.

    Args:
        seed (bytes, optional): if specified the keypair will be generated
            based on this seed, otherwise a random seed will be picked.
    """
    if seed is None:
        seed = nacl.utils.random(SEED_LEN)
    if len(seed) != SEED_LEN:
        raise ValueError("seed must contain %d bytes, got %s" % (SEED_LEN, len(seed)))

    signing_key = nacl.signing.SigningKey(seed)
    return signing_key


class Key:

    @staticmethod
    def from_signing_key(signing_key):
        """
        Args:
            signing_key (nacl.signing.SigningKey)

        Returns:
            Key
        """
        public = signing_key.verify_key
        seed = signing_key._seed
        return Key(public, signing_key, seed)

    @staticmethod
    def from_seed(seed=None):
        signing_key = generate_keypair(seed)
        return Key.from_signing_key(signing_key)

    @staticmethod
    def from_random():
        return Key.from_seed()

    @staticmethod
    def from_encoded_private(encoded_private):
        seed = from_pkcs8(encoded_private)
        return Key.from_seed(seed)

    def __init__(self, public, private, seed):
        encoded_public = public.encode()
        if len(encoded_public) != PUBLIC_KEY_LEN:
            raise ValueError(
                "Invalid public key, expected length %d got %d"
                % (PUBLIC_KEY_LEN, len(encoded_public))
            )

        encoded_private = private.encode()
        if len(encoded_private) != PRIVATE_KEY_LEN:
            raise ValueError(
                "Invalid private key, expected length %d got %d"
                % (PRIVATE_KEY_LEN, len(encoded_private))
            )

        if len(seed) != SEED_LEN:
            raise ValueError(
                "Invalid seed, expected length %d got %d" % (SEED_LEN, len(seed))
            )

        self.public = public
        self.private = private
        self.seed = seed
        self.encoded_public = to_x509(public.encode())
        self.encoded_private = to_pkcs8(private.encode())

        if len(self.encoded_public) != ENCODED_PUBLIC_KEY_LEN:
            raise ValueError("Unexpected encoding mismatch for public key")
        if len(self.encoded_private) != ENCODED_PRIVATE_KEY_LEN:
            raise ValueError("Unexpected encoding mismatch for private key")

    def to_address(self, raw=False):
        """
        Return the Semux address derived from the public key.
        """
        h160 = hash160(self.encoded_public)
        if len(h160) != ADDRESS_LEN:
            raise ValueError("Unexpected encoding mismatch for address")
        if not raw:
            h160 = binascii.hexlify(h160)
        return h160

    def sign(self, message):
        """
        Sign a message.
        """
        signed_msg = self.private.sign(message)
        return Signature(signed_msg.signature, self.public.encode())



from binascii import hexlify
from semux.hash import hash256, hash160


def test_hash256():
    assert hexlify(hash256(b'')) == b'0e5751c026e543b2e8ab2eb06099daa1d1e5df47778f7787faab45cdf12fe3a8'
    assert hexlify(hash256(b'test')) == b'928b20366943e2afd11ebc0eae2e53a93bf177a4fcf35bcc64d503704e65e202'


def test_hash160():
    assert hexlify(hash160(b'')) == b'a13dcd79a9c0e088317b189c88641f4ab671ab15'
    assert hexlify(hash160(b'test')) == b'86e8402b7615f07a2acb2ef1f4a54d323bbede77'



import nacl.signing

from .hash import hash160
from .encoding import to_x509

__all__ = ["SIG_LEN", "Signature"]

S_LEN = 64
A_LEN = 32
SIG_LEN = 96  # S_LEN + A_LEN


class Signature:

    @staticmethod
    def from_bytes(data):
        if len(data) != SIG_LEN:
            raise ValueError(
                "Invalid data length, expected %d got %d" % (SIG_LEN, len(data))
            )
        return Signature(data[:S_LEN], data[S_LEN:])

    def __init__(self, signed_message, public_key):
        if len(signed_message) != S_LEN:
            raise ValueError(
                "Invalid signed message length, expected %d got %d"
                % (S_LEN, len(signed_message))
            )
        if len(public_key) != A_LEN:
            raise ValueError(
                "Invalid public key length, expected %d got %d"
                % (A_LEN, len(public_key))
            )

        self.signed_message = signed_message
        self.public_key = public_key

    def get_verifying_key(self):
        return nacl.signing.VerifyKey(self.public_key)

    def get_address(self):
        return hash160(to_x509(self.public_key))

    def to_bytes(self):
        return self.signed_message + self.public_key



from setuptools import setup, find_packages

setup(
    name='semux',
    version='0.0.1',
    url='https://github.com/gpip/semux-py',
    packages=find_packages(),
    install_requires=[
      'PyNaCl==1.2.1',
      'pyblake2==1.1.2',
    ]
)



__all__ = ["to_pkcs8", "from_pkcs8", "to_x509"]

OID_OLD = 0x64
OID_ED25519 = 0x70
OID_BYTE = 0xb
IDLEN_BYTE = 0x6


def to_pkcs8(data):
    total_len = 16 + len(data)

    result = bytearray(
        [
            # sequence
            0x30,
            total_len - 2,
            # version
            0x02,
            1,
            # no public key included
            0,
            # sequence, algorithm identifier
            0x30,
            5,
            # oid
            0x06,
            3,
            (1 * 40) + 3,
            0x65,
            OID_ED25519,
            # no params set
            # octet string
            0x04,
            2 + len(data),
            # curve
            0x04,
            len(data),
        ]
    )
    return bytes(result) + data


def from_pkcs8(encoded_data):
    total_len = 48
    id_len = 5

    oid = encoded_data[OID_BYTE]
    if oid == OID_OLD:
        total_len = 49
        id_len = 8
    elif oid == OID_ED25519:
        if encoded_data[IDLEN_BYTE] == 7:
            total_len = 50
            id_len = 7
    else:
        raise ValueError("Unknown encoding format")

    if len(encoded_data) != total_len:
        raise ValueError(
            "Unexpected data length, got %d expected %d"
            % (len(encoded_data), total_len)
        )

    # Decode
    e = encoded_data
    if (
        e[0] != 0x30
        or e[1] != (total_len - 2)
        or e[2] != 0x02
        or e[3] != 1
        or e[4] != 0
        or e[5] != 0x30
        or e[6] != id_len
        or e[7] != 0x06
        or e[8] != 3
        or e[9] != (1 * 40) + 3
        or e[10] != 0x65
    ):
        raise ValueError("Unsupported key spec")

    # e[11] was checked above already.

    if oid == OID_OLD:
        if e[12] != 0x0a or e[13] != 1 or e[14] != 1:
            raise ValueError("Unsupported key spec")
        indx = 15
    else:
        indx = 12
        if id_len == 7:
            if e[indx] != 0x05 or e[indx + 1] != 0:
                raise ValueError("Unsupported key spec")
            indx = 14

        # octet string for the data
        if e[indx] != 0x04 or e[indx + 1] != 2 + 32:
            raise ValueError("Unsupported key spec")
        indx += 2

    if e[indx] != 0x04 or e[indx + 1] != 32:
        raise ValueError("Unsupported key spec")

    return e[indx + 2 : indx + 2 + 32]


def to_x509(data):
    total_len = 12 + len(data)

    result = bytearray(
        [
            # sequence
            0x30,
            total_len - 2,
            # sequence, algorithm identifier
            0x30,
            5,
            # oid
            0x06,
            3,
            (1 * 40) + 3,
            0x65,
            OID_ED25519,
            # no params set
            # key
            0x03,
            1 + len(data),
            0,
        ]
    )
    return bytes(result) + data



import hashlib

from pyblake2 import blake2b

__all__ = ["hash256", "hash160"]


def hash256(data):
    return blake2b(data, digest_size=32).digest()


def hash160(data):
    return hashlib.new("ripemd160", hash256(data)).digest()
